
"""Returns the fetched data by crawling"""
from urllib.parse import urlparse, urljoin
from urllib.error import URLError
from urllib.error import HTTPError
from collections import deque
from tqdm import tqdm
import urllib.request
import re
import sys
import datetime
import time
import random
import ssl
import traceback
from multiprocessing.pool import Pool

total_urls_visited = 0

class KeyboardInterruptError(Exception): 
    pass


class Pycrawler(object):
    """ Class Pycrawler id made"""

    def __init__(self, start_url):  # called when object is initialized
        self.start_url = start_url
        self.internal_urls = set()
        self.external_urls = set()
        self.broken_links = set()
        self.external = set()
        self.new_urls = deque([])
        self.processed_urls = set()
        self.total = 0
        self.dic = {}
        self.Dict = {}
        self.Dict["Internal"] = {}
        self.Dict["broken"] = {}
        self.Dict["External"] = {}
        self.internallist = []
        self.externallist = []
        self.brokenlist = []

    def startu(self):
        """Start Crawling"""
        self.crawl(self.start_url)

    def get_response_time(self, link):
        """get response time of url"""
        start = time.time()
        urllib.request.urlopen(link)
        diff = time.time() - start
        responsetime = round(diff, 2)
        return responsetime

    def is_valid(self, url):
        tokens = [urllib.parse.urlparse(url)]
        min_attributes = ("scheme", "netloc")  # add attrs to your liking
        for token in tokens:
            if not all([getattr(token, attr) for attr in min_attributes]):
                error = "'{url}' string has no scheme or netloc.".format(
                    url=token.geturl()
                )
                print(error)
            else:
                return True
            return False

    def get_all_website_links(self, url):
        """get all website links for the url passed"""
        self.urls = set()
        domain = urlparse(url).netloc
        try:
            response = urllib.request.urlopen(
                url, context=ssl._create_unverified_context()
            )
            responsetime = self.get_response_time(url)
            self.Dict["Internal"][url] = responsetime
            with response as resp:
                data = resp.read().decode("latin-1")
        except HTTPError as error:
            if error.code == 429:
                print("Too many requests, wait for sometime to continue the execution")
                time.sleep(360)
                print("Total External links:", len(self.external_urls))
                print(" Total Internal links:", len(self.internal_urls))
                total = (
                    len(self.external_urls)
                    + len(self.internal_urls)
                    + len(self.broken_links)
                )
                print("Total links:", self.total)
                return None
            else:
                if url in self.internal_urls:
                    if url not in self.broken_links:
                        self.broken_links.add(url)
                        self.Dict["broken"][url] = str(error)[11:]
                    self.internal_urls.remove(url)
                return None
        except (URLError, ValueError) as error:
            time.sleep(5)
            traceback.print_exc()
            if url in self.internal_urls:
                if url not in self.broken_links:
                    self.broken_links.add(url)
                    self.Dict["broken"][url] = str(error)[11:]
                self.internal_urls.remove(url)
            return None
        links = re.findall(
            '''<a\s+(?:[^>]*?\s+)?href="([^"]*)"''', data
        )  # Extracting the matching urls
        rlinks = re.findall('''<img\s+(?:[^>]*?\s+)?src="([^"]*)"''', data)
        all_links = links + rlinks
        for link in tqdm(all_links):
            href = link
            if href == "" or href is None:
                continue
            href = urljoin(url, href)
            parsed_href = urlparse(href)
            # remove URL GET parameters, URL fragments, etc.
            href = parsed_href.scheme + "://" + parsed_href.netloc + parsed_href.path
            if not self.is_valid(href):  # not a valid URL
                continue
            if href in self.internal_urls:
                continue  # already in the set
            if href in self.broken_links:
                continue
            if domain not in href:  # external link
                if href not in self.external_urls:
                    self.external_urls.add(href)
                continue
            self.urls.add(href)
            self.internal_urls.add(href)
        return self.urls  # return internal_urls to be crawled

    def crawl(self, url):
        """
        Crawls a web page and extracts all links.
        You'll find all links in `external_urls` and `internal_urls` global set variables.
        params:
        """
        if url == self.start_url:
            print(f"Crawling {url} ...")
        self.new_urls.append(url)
        while len(self.new_urls) > 0:
            link = self.new_urls[0]
            self.new_urls.popleft()
            try:
                time.sleep(3)
                global total_urls_visited
                total_urls_visited += 1
                if total_urls_visited % 10 == 0:
                    print(f"Total No of links crawled so far={total_urls_visited}")
                links = self.get_all_website_links(link)
                self.processed_urls.add(link)
                if links:
                    for each in links:
                        self.new_urls.append(each)
            except (RuntimeError, TypeError, NameError):
                time.sleep(2)
                traceback.print_exc()
            except:
                print("Unexpected error:", sys.exc_info()[0])
                raise


def mainfunc(webname):
    try:
        """All the main implementation of crawling and presenting information is handled"""
        crawler = Pycrawler(webname)
        crawler.startu()
        time.sleep(10)
        # remove broken links from external_links
        for link in crawler.external_urls:
            try:
                responsetime = crawler.get_response_time(link)
                crawler.external.add(link)
                crawler.Dict["External"][link] = responsetime
            except (HTTPError, URLError) as error:
                crawler.broken_links.add(link)
                crawler.Dict["broken"][link] = str(error)[11:]
        crawler.total = (
            len(crawler.external) + len(crawler.internal_urls) + len(crawler.broken_links)
        )
        print(f"Searched website: {crawler.start_url}")
        if crawler.total > 0:
            print(
                f"Found {crawler.total-len(crawler.broken_links)} links ("
                + str(len(crawler.external))
                + " external references, "
                + str(len(crawler.internal_urls))
                + " internal references)"
            )
            now = datetime.datetime.now()
            dt_string = now.strftime("%d/%m/%Y %H:%M:%S IST")
            string = "abcdefghijklmnopqrstuvwxyz0123456789"
            reportid = "".join(random.sample(string, 6))
            crawler.dic = {
                "report-id": reportid,
                "starting_url": crawler.start_url,
                "External links": crawler.external,
                "Internal links": crawler.internal_urls,
                "Broken links": crawler.broken_links,
                "Date and Time": dt_string,
                "Total links": crawler.total,
            }
            print("Report generated with unique-id:", reportid)
            del crawler.Dict["Internal"][crawler.start_url]
            for i in crawler.Dict["Internal"].items():
                webinfo = []
                webinfo.append(i[0])
                webinfo.append(str(i[1]) + "s")
                crawler.internallist.append(list(webinfo))
            for i in crawler.Dict["External"].items():
                webinfo = []
                webinfo.append(i[0])
                webinfo.append(str(i[1]) + "s")
                crawler.externallist.append(list(webinfo))
            for i in crawler.Dict["broken"].items():
                webinfo = []
                webinfo.append(i[0])
                webinfo.append(i[1])
                crawler.brokenlist.append(list(webinfo))
            data = {}
            data["report_id"] = crawler.dic["report-id"]
            data["website_link"] = crawler.dic["starting_url"]
            data["total_links"] = crawler.dic["Total links"]
            data["external_links"] = crawler.externallist
            data["internal_links"] = crawler.internallist
            data["broken_links"] = crawler.brokenlist
            data["date_time"] = crawler.dic["Date and Time"]
            return data
        else:
            print("Invalid URL or No Links Found")
        
    except KeyboardInterrupt:
        print("Ctrl + c Captured")
        print()
        raise KeyboardInterruptError()



def concurrency(ipurls):
    with Pool(4) as tasks:
        somelist = tasks.map(mainfunc,ipurls)
    return somelist

if __name__=="__main__":
    concurrency(["https://httpbin.org/","https://erp.aktu.ac.in/webpages/oneview/oneview.aspx"])
    


import sys
try:
    from . import storage
    from . import representation
    from . import helpcode
    from . import crawler
except:
    import storage
    import representation
    import helpcode
    import crawler

import argparse

def cmd_execute():
    parser = argparse.ArgumentParser("Stats about websites you need to Crawl ")
    parser.add_argument('commands', type=str, nargs='*')
    args = parser.parse_args()

    store = storage.Store()
    con = store.connect_db()
    store.make_table(con)

    if len(args.commands) > 0:
        if args.commands[0] == "crawl" and len(args.commands)>1:
            ipurls = []
            for i in range(1, len(args.commands)):
                ipurls.append(args.commands[i])
            data = crawler.concurrency(ipurls)
            for i in range(len(data)):
                if data[i] is not None:
                    store.insert(con, data[i])
        elif args.commands[0] == "list-reports" and len(args.commands) == 1:
            reports = store.fetch(con, "list-reports")
            iteration = 1
            for i in reports:
                print(
                    str(iteration)
                    + ". "
                    + reports[i][0]
                    + " - "
                    + reports[i][1]
                    + " - "
                    + reports[i][2]
                )
                iteration += 1

        elif args.commands[0] == "links-report" and 1 < len(args.commands) < 4:
            report = store.fetch(con, "links-report", args.commands[1])
            if report is not None:
                if len(args.commands) == 2:
                    print(representation.main(report))
                elif args.commands[2] == "json" or args.commands[2] == "csv" or args.commands[2] == "html":
                    print(representation.main(report, args.commands[2]))
                else:
                    print("Wrong format selected")
                    print("Available Formats - YAML, HTML, JSON, CSV")
                    helpcode.help_func()

        else:
            helpcode.help_func()
    else:
        helpcode.help_func()
    store.close_dbconnect(con)
    
if __name__=="__main__":
    cmd_execute()


data = {
    "report_id": "4455bf",
    "website_link": "https://www.an-old-website.com/",
    "total_links": 17,
    "internal_links": [
        ["/about", "0.34s"],
        ["/contact", "0.23s"],
        ["/blog", "0.56s"],
        ["/blog/2020/07/12/why-is-python-preferred-for-automation.html", "0.72s"],
        ["/blog/2020/07/15/is-nim-a-python-killer.html", "0.66s"],
        ["/blog/2020/07/16/python-and-javascript-best-of-both-worlds.html", "0.89s"],
        ["/favicon.ico", "0.22s"],
        ["/assets/images/logo.jpg", "0.34s"],
        ["/assets/images/banner.jpg", "0.56s"],
        ["/assets/css/main.css", "0.33s"],
    ],
    "external_links": [
        ["https://ajax.googleapis.com/ajax/libs/jquery/3.5.1/jquery.min.js", "1.23s"],
        [
            "https://maxcdn.bootstrapcdn.com/bootstrap/3.4.1/js/bootstrap.min.js",
            "0.97s",
        ],
        [
            "https://maxcdn.bootstrapcdn.com/bootstrap/3.4.1/css/bootstrap.min.css",
            "0.92s",
        ],
        ["https://search.google.com", "1.07s"],
    ],
    "broken_links": [
        [
            "/blog/2020/07/10/test-automation-techniques-using-python.html",
            "404_NOT_FOUND",
        ],
        ["https://www.faecbook.com/user/ElectricBoa", "404_NOT_FOUND"],
        ["../pelican-installation-procedure.html", "404_NOT_FOUND"],
    ],
    "data_time": "2019/12/22 11:30:00 IST",
}




"""Help module of command line"""

def help_func():
    """help options"""
    print("\n Usage: \n")
    print("\t python website_stats.py [command] [Argument] \n ")
    print("Available Commands: \n")
    print(
        "\t crawl -> This command collects all the statistics of a website passed as argument."
    )
    print("\t \t - it takes a valid website URL as argument. \n")
    print(
        "\t list-reports -> This command lists the report snapshot of each crawled website."
    )
    print(
        "\t \t    - it does not take any argument, just lists all the crawled websites \n"
    )
    print(
        """\t links-report -> This command lists all the links report of a specific website crawl operation."""
    )
    print("\t \t      - This command takes the report id as the argument. \n")




"""Generation of Report in different forms"""
import json
import csv
import yaml
import os
import webbrowser


class Representation:
    """generate report in different format"""

    def __init__(self, data):
        self.data = data

    def selecttype(self, representation_format=None):
        """select format"""
        if representation_format == "json":
            return self.json_report()
        if representation_format == "csv":
            return self.csv_report()
        if representation_format == "html":
            return self.html_report()
        return self.yaml_report()

    def dict_to_html_table(self):
        tbl_fmt = """
        <table>{}
        </table>"""

        row_fmt = """
        <tr>
            <td>{}</td>
            <td>{}</td>
        </tr>"""
        return tbl_fmt.format(
            "".join([row_fmt.format(k, v) for k, v in self.data.items()])
        )

    def yaml_report(self):
        """report in yaml"""
        yaml_data = yaml.dump(self.data)
        # print(self.yaml_data)
        yaml_dte = ""
        yaml_data1 = yaml.load_all(yaml_data, Loader=yaml.FullLoader)
        for doc in yaml_data1:
            yaml_dte += "report-id: " + doc["report-id"] + "\n"
            yaml_dte += "website: " + doc["website"] + "\n"
            yaml_dte += "total-links: " + str(doc["total-links"]) + "\n\n"
            yaml_dte += "external-references: " + "\n"
            yaml_dte += (
                "   # [ links-to-external-websites, response-time-in-seconds ] \n"
            )
            for j in range(len(doc["external-references"])):
                yaml_dte += (
                    "   - [ "
                    + doc["external-references"][j][0]
                    + ", "
                    + doc["external-references"][j][1]
                    + " ]\n"
                )
            yaml_dte += "\n"
            yaml_dte += "internal-references: " + "\n"
            yaml_dte += "   # [ links-to-internal-webpages-and-resources, response-time-in-seconds ]\n"
            for j in range(len(doc["internal-references"])):
                yaml_dte += (
                    "   - [ "
                    + doc["internal-references"][j][0]
                    + ", "
                    + doc["internal-references"][j][1]
                    + " ]\n"
                )
            yaml_dte += "\n"
            yaml_dte += "broken-links: " + "\n"
            yaml_dte += (
                "   # [ links-that-could-not-be-resolved, http-error-status-code ]\n"
            )
            for j in range(len(doc["broken-links"])):
                yaml_dte += (
                    "   - [ "
                    + doc["broken-links"][j][0]
                    + ", "
                    + doc["broken-links"][j][1]
                    + " ]\n"
                )

        return yaml_dte

    # html_report_generation
    def html_report(self):
        x = self.dict_to_html_table()
        filename = "reportgenerated.html"
        print(x)
        f = open(filename, "w")
        f.write(x)
        f.close()
        return webbrowser.open("file://" + os.path.realpath(filename))

    # json_report_generation
    def json_report(self):
        """report in json"""
        json_data = json.dumps(self.data, indent=2)
        return json_data

    # csv_report_generation
    def csv_report(self):
        """report in csv"""
        with open("./csv_report.csv", "w") as csv_file:
            writer = csv.writer(csv_file)
            for key, value in self.data.items():
                writer.writerow([key, value])
        doc = open("./csv_report.csv", "r")
        return doc.read()


def main(data, representation_format=None):
    """return format data"""
    representation = Representation(data)
    return representation.selecttype(representation_format)



"""Functions of database to handle database"""
import sqlite3
from sqlite3 import Error
import ast


class Store:
    def __init__(self):
        pass

    def gen_report(self, k):
        """report generation"""
        reports = {}
        listheader = [
            "report-id",
            "website",
            "total-links",
            "external-references",
            "internal-references",
            "broken-links",
        ]
        for i in range(len(listheader)):
            reports[listheader[i]] = k[0][i]

        reports["external-references"] = ast.literal_eval(
            reports["external-references"]
        )
        reports["internal-references"] = ast.literal_eval(
            reports["internal-references"]
        )
        reports["broken-links"] = ast.literal_eval(reports["broken-links"])
        return reports

    def connect_db(self):
        """Database Connection"""
        try:
            con = sqlite3.connect("webstats2.db")
            return con

        except Error:
            print(Error)

    def make_table(self, con):
        """make table for storing website"""
        try:
            cur = con.cursor()
            cur.execute(
                """CREATE TABLE IF NOT EXISTS WEB(
                        Reportid VARCHAR(6) PRIMARY KEY NOT NULL,
                        Website VARCHAR(255) NOT NULL,
                        Totallinks INTEGER,
                        EXTERNAL VARCHAR(255),
                        INTERNAL VARCHAR(255),
                        BROKEN VARCHAR(255),
                        datetime VARCHAR(50) NOT NULL)
                        """
            )
            con.commit()
        except Error as error:
            print(error)

    def insert(self, con, data):
        """insert website data"""
        try:
            cur = con.cursor()

            webdata = [
                (
                    data["report_id"],
                    data["website_link"],
                    data["total_links"],
                    str(data["external_links"]),
                    str(data["internal_links"]),
                    str(data["broken_links"]),
                    data["date_time"],
                )
            ]
            cur.executemany("Insert Into WEB Values (?,?,?,?,?,?,?)", webdata)
            con.commit()

        except Error as error:
            print(error)

    def fetch(self, con, arg, reportid=None):
        """fetch data"""
        try:
            reports = {}
            cur = con.cursor()
            # fetch list reports
            if arg == "list-reports" and reportid is None:
                for report in cur.execute("SELECT Reportid,datetime,Website FROM WEB;"):
                    reports[report[0]] = report
                return reports
            # fetch report by report id
            elif arg == "links-report" and reportid is not None:
                try:
                    cur.execute(
                        """SELECT Reportid, Website, Totallinks,EXTERNAL,INTERNAL,BROKEN FROM WEB
                        WHERE Reportid ='%s'"""
                        % reportid
                    )
                    row = cur.fetchall()
                    reports = self.gen_report(row)
                    return reports

                except:
                    print("Report Not Found!")
                    return None
            else:
                print("Invalid Command Given")
                return None
        except Error as error:
            print(error)
            return None

    def close_dbconnect(self, con):
        """close connection"""
        con.close()
